import numpy as np
import sympy as sp
from gcoreutils.nsarray import NSArray
from multipie.data.data_transform_matrix import _data_trans_Ah, _data_trans_lattice_p, _data_trans_lattice_t


# 3x3 matrix to convert from reduced to cartesian coordinate.
Ah = (NSArray(_data_trans_Ah), NSArray(_data_trans_Ah).inverse())

# 4x4 matrix to convert from conventioanl to primitive coordinate.
latticeP = {lat: (NSArray(d), NSArray(d).inverse()) for lat, d in _data_trans_lattice_p.items()}

# set of translational vectors.
latticeT = {lat: NSArray(d) for lat, d in _data_trans_lattice_t.items()}


# ==================================================
def to_cartesian(crystal, vm):
    """
    convert from reduced to cartesian coordinates.

    Args:
        vm (NSArray): vector/matrix/bond to convert.

    Returns:
        NSArray: converted array.
    """
    if crystal in ["hexagonal", "trigonal"]:
        return vm.transform(Ah[0], Ah[1])
    else:
        return vm


# ==================================================
def to_reduced(crystal, vm):
    """
    convert from cartesian to reduced coordinates.

    Args:
        vm (NSArray): vector/matrix/bond to convert.

    Returns:
        NSArray: converted array.
    """
    if crystal in ["hexagonal", "trigonal"]:
        return vm.transform(Ah[1], Ah[0])
    else:
        return vm


# ==================================================
def to_conventional(lattice, vm, plus_set=False):
    """
    convert from primitive to conventional cells in reduced coordinate.

    Args:
        lattice (str): crystal lattice (A/B/C/P/I/F/R).
        vm (NSArray): vector/matrix/bond to convert.
        plus_set (bool, optional): when True, add partial translations.

    Returns:
        NSArray: converted array. when plus_set is True, return [set(t0),set(t1),...]
    """
    if lattice == "P":
        return vm
    else:
        P = latticeP[lattice]
        vmP = vm.transform(P[0], P[1])
        if plus_set:
            tl = latticeT[lattice]
            lst = []
            if vm.style == "matrix":
                for tli in tl:
                    A = vmP.copy()
                    A[0:3, 3] = tli
                    lst.append(A)
            else:
                A = NSArray("[[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]]")
                for tli in tl:
                    A[0:3, 3] = tli
                    lst.append(vmP.transform(A))
            return NSArray.concat(lst)
        else:
            return vmP


# ==================================================
def to_primitive(lattice, vm):
    """
    convert from conventional to primitive cells in reduced coordinate.

    Args:
        lattice (str): crystal lattice (A/B/C/P/I/F/R).
        vm (NSArray): vector/matrix/bond to convert.

    Returns:
        NSArray: converted array.
    """
    if lattice == "P":
        return vm
    else:
        P = latticeP[lattice]
        return vm.transform(P[1], P[0])


# ==================================================
def to_axial(mp):
    """
    convert from matrix for polar vector to one for axial vector (cartesian).

    Args:
        mp (NSArray): transform matrix for polar vector (3x3).

    Returns:
        NSArray: transform matrix for axial vector (3x3).
    """

    def axial(m):
        return np.array(
            [
                [
                    m[1][1] * m[2][2] - m[1][2] * m[2][1],
                    m[1][2] * m[2][0] - m[1][0] * m[2][2],
                    m[1][0] * m[2][1] - m[1][1] * m[2][0],
                ],
                [
                    m[0][2] * m[2][1] - m[0][1] * m[2][2],
                    m[0][0] * m[2][2] - m[0][2] * m[2][0],
                    m[0][1] * m[2][0] - m[0][0] * m[2][1],
                ],
                [
                    m[0][1] * m[1][2] - m[0][2] * m[1][1],
                    m[0][2] * m[1][0] - m[0][0] * m[1][2],
                    m[0][0] * m[1][1] - m[0][1] * m[1][0],
                ],
            ]
        )

    return NSArray(NSArray._apply(axial, mp, "matrix"), "matrix")


# ==================================================
def rotation_matrix(n, axis):
    """
    rotation matrix (3x3).

    Args:
        n (int): n-fold rotation (1, 2, 3, 4, 6, -1, -3, -4, -6).
        axis (NSArray): rotation axis (normalization free).

    Returns:
        NSArray: rotatoin matrix (3x3).
    """
    th = 2 * sp.pi / n
    a = axis.normalize()

    c = sp.cos(th)
    s = sp.sin(th)
    c1 = 1 - c

    m3 = sp.Matrix(
        [
            [c + a[0] * a[0] * c1, a[0] * a[1] * c1 - a[2] * s, a[2] * a[0] * c1 + a[1] * s],
            [a[0] * a[1] * c1 + a[2] * s, c + a[1] * a[1] * c1, a[1] * a[2] * c1 - a[0] * s],
            [a[2] * a[0] * c1 - a[1] * s, a[1] * a[2] * c1 + a[0] * s, c + a[2] * a[2] * c1],
        ]
    )
    m3 = NSArray(m3.tolist(), "matrix")

    return m3


# ==================================================
def reflection_matrix(axis):
    """
    reflection matrix (3x3).

    Args:
        axis (NSArray): reflection axis (normalization free).

    Returns:
        NSArray: reflection matrix (3x3).
    """
    return -rotation_matrix(2, axis)


# ==================================================
def rotoinversion_matrix(n, axis):
    """
    rotoinversion matrix (3x3).

    Args:
        n (int): n-fold rotation.
        axis (NSArray): rotaton axis (normalization free).

    Returns:
        NSArray: rotoinversion matrix (3x3).
    """
    return -rotation_matrix(n, axis)



"""
This file provides utility functions for tag.
"""
import copy
from dataclasses import asdict


# ==================================================
class TagList(list):
    """
    list of tags.
    """

    # ==================================================
    def __init__(self, lst=None):
        """
        create tags.

        Args:
            lst ([Tag]): list of tag class.
        """
        if lst is None:
            lst = []
        elif not hasattr(lst, "__iter__"):
            lst = [lst]
        super().__init__(lst)

    # ==================================================
    def select(self, **kwargs):
        """
        select tags with given keywords.

        Args:
            kwargs (dict): field-value pairs to select.

        Returns:
            TagList: selected tag list.
        """
        key = kwargs.items()
        x = [i for i in self if (asdict(i).items() & key) == set(key)]
        return TagList(x)

    # ==================================================
    def latex(self):
        """
        convert to list of LaTeX.

        Returns:
            [str] or str: latex string.
        """
        return [i.latex() for i in self]

    # ==================================================
    def symbol(tags, commutative=False):
        """
        convert to list of non-commutative symbol.

        Returns:
            [sympy]: non-commutative symbol.
        """
        return [i.symbol(commutative=commutative) for i in tags]

    # ==================================================
    def str_list(tags):
        """
        convert to list of string.

        Returns:
            [str] or str: tag string.
        """
        return [str(i) for i in tags]

    # ==================================================
    @classmethod
    def from_str(self, tag_type, tag_list):
        """
        create tags.

        Args:
            tag_type (Tag): Tag class.
            tag_list ([str]): tag list in str.
        """
        return TagList([tag_type(tag) for tag in tag_list])

    # ==================================================
    def __getslice__(self, start, stop):
        lst = super().__getslice__(start, stop)
        return self.__class__(lst)

    # ==================================================
    def __getitem__(self, key):
        if isinstance(key, slice):
            lst = super().__getitem__(key)
            return self.__class__(lst)
        else:
            return super().__getitem__(key)

    # ==================================================
    def __iadd__(self, other):
        if isinstance(other, TagList):
            for i in other:
                self.append(i)
        return self

    # ==================================================
    def __add__(self, other):
        s = copy.copy(self)
        s += other
        return s



"""
SymmetryOperationG manages point/space-group symmetry operations.
"""
import sympy as sp
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_list import TagList
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_symmetry_operation import TagSymmetryOperation
from multipie.symmetry_operation.util.symmetry_operation_util import latticeT, to_cartesian, to_primitive, to_conventional
from multipie.symmetry_operation.symmetry_operation import SymmetryOperation
from multipie.data.data_symmetry_operation_pg import _data_symmetry_operation_pg
from multipie.data.data_symmetry_operation_sg import _data_symmetry_operation_sg


# ==================================================
class SymmetryOperationG(
    dict
):  # dict of (symmetry operation tag, symmetry operation), {TagSymmetryOperation: SymmetryOperation}.
    """
    point/space-group symmetry operations.

    Attributes:
        tag (TagGroup): point-group tag.
        gen ([TagSymmetryOperation]): generators.
        cc ([[TagSymmetryOperation]]): symmetry operations categorized by conjugacy class.
        cc1 ([TagSymmetryOperation]): representative symmetry operations in conjugacy class.
        full ([TagSymmetryOperation]): all symmetry operations.
        cc_num ([int]): the number of symmetry operations in conjugacy class.
        crystal (str): crystal type, (triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic).
        lattice (str): lattice type, (A/B/C/P/I/F/R/null).
        translation (NSArray): partial translation vectors (reduced coordinate).
    """

    #
    #     __idx_cc1 ([int]): index of top of conjugacy class.
    #     __product ({(TagSymmetryOperation,TagSymmetryOperation): TagSymmetryOperation}): product table (point-group only).
    #     __inv ({TagSymmetryOperation:TagSymmetryOperation}): inverse operation.
    #
    # ==================================================
    def __init__(self, g_tag):
        """
        initialize the class.

        Args:
            g_tag (TagGroup or str): tag of point/space group.
        """
        g_tag = TagGroup(str(g_tag))
        self.tag = g_tag
        """point-group tag."""

        if self.tag.is_point_group():
            gen, cc = _data_symmetry_operation_pg[str(g_tag)]
        else:
            gen, cc = _data_symmetry_operation_sg[str(g_tag)]

        self.gen = TagList.from_str(TagSymmetryOperation, gen)
        """generators."""

        self.cc = [TagList.from_str(TagSymmetryOperation, i) for i in cc]
        """symmetry operations categorized by conjugacy class."""

        self.cc1 = TagList.from_str(TagSymmetryOperation, [i[0] for i in cc])
        """representative symmetry operations in conjugacy class."""

        self.cc_num = [len(i) for i in self.cc]
        """the number of symmetry operations in conjugacy class."""

        self.full = TagList.from_str(TagSymmetryOperation, sum(cc, []))
        """all symmetry operations."""

        self.__idx_cc1 = [self.full.index(i) for i in self.cc1]

        self.crystal = self.tag.crystal
        """crystal type, (triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic)."""

        self.lattice = self.tag.lattice
        """lattice type, (A/B/C/P/I/F/R/null)."""

        if self.lattice != "":
            self.translation = latticeT[self.lattice]
            """partial translation vectors (reduced coordinate)."""
        else:
            self.translation = None

        for s in self.full:
            self[s] = SymmetryOperation(str(s), self.crystal)

        # product table.
        self.__product = {}
        self.__inv = {}
        if self.tag.is_point_group():
            so = self.mat()
            so_inv = so.inverse()
            t = self.key_list()
            for i in range(len(so)):
                soi = so[i]
                self.__inv[t[i]] = t[so.index(so_inv[i])]
                for j in range(len(so)):
                    soj = so[j]
                    self.__product[(t[i], t[j])] = t[so.index(soi @ soj)]

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    @property
    def plus_set(self):
        """
        additional partial translation vectors in space group.

        Returns:
            NSArray: translation vectors.
        """
        assert not self.tag.is_point_group(), "plus_set is only for space group."
        return self.translation

    # ==================================================
    def mat(self, axial=False, cc_only=False, primitive=False, cartesian=False):
        """
        matrix for symmetry operatoins.

        Args:
            axial (bool, optional): matrix for axial vector ?
            cc_only (bool, optional): conjugacy-class operations only ?
            primitive (bool, optional): for primitive cell ? (space group only)
            cartesian (bool, optional): in cartesian coordinate ?

        Returns:
            list: list of operations (3x3: point group) or (4x4: space group), [NSArray].
        """
        if cc_only:
            idx = TagList([self.full[i] for i in self.__idx_cc1])
        else:
            idx = self.full
        if axial:
            ops = NSArray([self[i].m_axial.tolist() for i in idx], "matrix")
        else:
            ops = NSArray([self[i].m_polar.tolist() for i in idx], "matrix")
        if primitive:
            ops = to_primitive(self.lattice, ops)
        if cartesian:
            ops = to_cartesian(self.crystal, ops)

        return ops

    # ==================================================
    def __getitem__(self, item):
        if type(item) == str:
            return self.get(TagSymmetryOperation(item))
        else:
            return self.get(item)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())

    # ==================================================
    def inverse(self, i):
        """
        inverse of symmetry operation (point group only).

        Args:
            i (TagSymmetryOperation): symmetry operation.

        Returns:
            TagSymmetryOperation: inverse of symmetry operation.
        """
        assert self.tag.is_point_group(), "product table is available only for point group."

        if type(i) == str:
            i = TagSymmetryOperation(i)
        return self.__inv[i]

    # ==================================================
    def product(self, i, j):
        """
        product of symmetry operations (point group only).

        Args:
            i (TagSymmetryOperation): symmetry operation (left).
            j (TagSymmetryOperation): symmetry operation (right).

        Returns:
            TagSymmetryOperation: product of symmetry operations.
        """
        assert self.tag.is_point_group(), "product table is available only for point group."

        if type(i) == str:
            i = TagSymmetryOperation(i)
        if type(j) == str:
            j = TagSymmetryOperation(j)
        return self.__product[(i, j)]

    # ==================================================
    def _transform_matrix(self, phi_p, phi, vs):
        """
        transform matrix for polynomial basis (phi_i' = T_ij phi_j).

        Args:
            phi_p (NSArray): transformed polynomials.
            phi (NSArray): polynomials to be transformed.
            vs (NSArray): list of variables in polynomials.

        Returns:
            NSArray: transform matrix (phi' = M.phi).
        """
        cv = NSArray(list(sp.symbols(f"coeffvar0:{phi.size}")), "vector")
        eqs = [phi_p[i].tolist() - phi @ cv for i in range(phi_p.size)]
        eqs = [i.expand() for i in eqs]
        tm = sp.conjugate(sp.Matrix([list(sp.linsolve(sp.Poly(eq, vs.tolist()).coeffs(), cv.tolist()))[0] for eq in eqs]))

        return NSArray(tm.tolist(), "matrix")

    # ==================================================
    def _transform_variable(self, phi, v_old, v_new):
        """
        transform polynomials by substituting transformed variables.

        Args:
            phi (NSArray): polynomials to be transformed.
            v_old (NSArray): list of variables in polynomials.
            v_new (NSArray): list of transformed variables.

        Returns:
            NSArray: transformed polynomial or list of polynomials.
        """
        return phi.subs({v_old[i]: v_new[i] for i in range(v_old.size)})

    # ==================================================
    def _transform_matrix_orbital_basis(self, v, vp, phi, pdet=None):
        """
        transform matrix for polynomial basis, (phi_i' = T_ij phi_j).

        Args:
            v (NSArray): polar vector variable (cartesian coordinate).
            vp (NSArray): list of transformed polar vectors, r' (cartesian coordinate).
            phi (NSArray): list of basis (cartesian coordinate).
            pdet (NSArray, optional): list of determinant.

        Returns:
            NSArray: list of transform matrices.
        """
        tm = []
        if pdet is None:
            for i in range(len(vp)):
                phi_p = self._transform_variable(phi, v, vp[i])
                tm.append(self._transform_matrix(phi_p, phi, v).tolist())
        else:
            for i in range(len(vp)):
                phi_p = self._transform_variable(phi, v, vp[i])
                phi_p = [phi_p[j] * pdet[j] for j in range(len(phi_p))]
                tm.append(self._transform_matrix(phi_p, phi, v).tolist())

        return NSArray(tm, "matrix")

    # ==================================================
    def _equivalent_vector(
        self, v, cc_only=False, axial=False, remove_duplicate=False, cartesian=False, primitive=False, plus_set=False, shift=False
    ):
        """
        create symmetry-operation equivalent vectors.

        Args:
            v (str or NSArray): vector/site position, "[x,y,z]".
            cc_only (bool, optional): conjugacy-class operations only ?
            axial (bool, optional): axial vector ?
            remove_duplicate (bool, optional): remove the same sites ?
            cartesian (bool, optional): cartesian coordinate ? (point group only).
            primitive (bool, optional): primitive cell ? (space group only).
            plus_set (bool, optional): add partial translations ? (space group only).
            shift (bool, optional): move into unit cell ? (space group only).

        Returns:
            NSArray: equivalent vectors.

        Notes:
            - if remove_duplicate is True, return values are sorted.
        """
        if not isinstance(v, (str, NSArray)):
            raise KeyError(f"{type(v)} is not accepted for vector/site.")

        if type(v) == str:
            v = NSArray(v)

        if self.tag.is_point_group():
            ops = self.mat(axial=axial, cc_only=cc_only, cartesian=cartesian)

            ev = ops.apply(v)
            if remove_duplicate:
                ev = ev.remove_duplicate()
        else:
            ops = self.mat(axial=axial, cc_only=cc_only)
            ops = to_primitive(self.lattice, ops)
            v = to_primitive(self.lattice, v)
            ev = ops.apply(v)

            if shift:
                ev = ev.shift()

            if remove_duplicate:
                ev = ev.remove_duplicate().sort()

            if not primitive:
                ev = to_conventional(self.lattice, ev, plus_set=plus_set)
                if type(ev) == list:
                    ev = NSArray.concat(ev)
                if shift:
                    ev = ev.shift()
                if remove_duplicate:
                    ev = ev.remove_duplicate().sort()

        return ev

    # ==================================================
    def _equivalent_bond(
        self,
        bond,
        cc_only=False,
        remove_duplicate=False,
        cartesian=False,
        primitive=False,
        plus_set=False,
        shift=False,
        nondirectional=False,
    ):
        """
        create SO-equivalent bonds.

        Args:
            bond (str or NSArray): bond, "vector@center", "tail;head", "start:vector".
            cc_only (bool, optional): conjugacy-class operations only ?
            remove_duplicate (bool, optional): remove the same bonds ?
            cartesian (bool, optional): cartesian coordinate ? (point group only).
            primitive (bool, optional): primitive cell ? (space group only).
            plus_set (bool, optional): add partial translations ? (space group only).
            shift (bool, optional): move into unit cell ? (space group only).
            nondirectional (bool, optional): ignore directional property ?

        Returns:
            NSArray: equivalent bonds.

        Notes:
            - if bond is string, vector-center "[vx,vy,vz]@[cx,cy,cz]", tail-head "[tx,ty,tz];[hx,hy,hz], or start-vector "[sx,sy,sz]:[vx,vy,vz]" style is accepted.
            - if remove_duplicate is True, return values are sorted.
        """
        if not isinstance(bond, (str, NSArray)):
            raise KeyError(f"{type(bond)} is not accepted for bond.")

        if type(bond) == str:
            bond = NSArray(bond)

        t, h = bond.convert_bond("bond_th")
        if self.tag.is_point_group():
            t = self._equivalent_vector(t, cc_only=cc_only, cartesian=cartesian)
            h = self._equivalent_vector(h, cc_only=cc_only, cartesian=cartesian)
            b = NSArray.create_bond_from_pair(t, h, "bond_th")
            if nondirectional:
                b = b.regular_direction()
            if remove_duplicate:
                b = b.remove_duplicate().sort()
        else:
            t = self._equivalent_vector(t, cc_only=cc_only)
            h = self._equivalent_vector(h, cc_only=cc_only)
            t = to_primitive(self.lattice, t)
            h = to_primitive(self.lattice, h)
            b = NSArray.create_bond_from_pair(t, h, "bond_th")
            if shift:
                b = b.shift()
            if nondirectional:
                b = b.regular_direction()
            if remove_duplicate:
                b = b.remove_duplicate().sort()

            if not primitive:
                t, h = b.convert_bond("bond_th")
                t = to_conventional(self.lattice, t, plus_set=plus_set)
                h = to_conventional(self.lattice, h, plus_set=plus_set)
                b = NSArray.create_bond_from_pair(t, h, "bond_th")
                if shift:
                    b = b.shift()
                if nondirectional:
                    b = b.regular_direction()
                if remove_duplicate:
                    b = b.remove_duplicate().sort()

        return b



"""
CharacterPG manages point-group character table.
"""
import sympy as sp
from gcoreutils.convert_util import text_to_sympy
from multipie.tag.tag_group import TagGroup
from multipie.tag.tag_irrep import TagIrrep
from multipie.tag.tag_symmetry_operation import TagSymmetryOperation
from multipie.tag.tag_list import TagList
from multipie.data.data_character_table import _data_character
from multipie.data.data_product_decomp import _data_product_decomp
from multipie.data.data_compatibility_relation import _data_compatibility_relation


# ==================================================
class CharacterPG(dict):  # dict of (irrep, characters), {TagIrrep: [sympy]}.
    """
    point-group character table.

    Attributes:
        tag (TagGroup): point-group tag.
    """

    #
    #     __irrep (list): irrep., [TagIrrep].
    #     __symmetry_operation (list): symmetry operations of conjugacy class, [TagSymmetryOperation].
    #     __so_number (list): number of SOs in each conjugacy class, [int].
    #     __character_symbol (dict): character table in symbolic form, {TagIrrep:[sympy]}.
    #     __parity_conv (dict): parity conversion, {TagIrrep: TagIrrep}.
    #     __product_decomp_s (dict): symmetric product decomposition, {(TagIrrep,TagIrrep):[(int,TagIrrep)]}.
    #     __product_decomp_a (dict): anti-symmetric product decomposition, {TagIrrep:[(int,TagIrrep)]}.
    #
    # ==================================================
    def __init__(self, pg_tag):
        """
        initialize the class.

        Args:
            pg_tag (TagGroup or str): point-group tag.
        """
        pg_tag = TagGroup(str(pg_tag))

        # set tag.
        self.tag = pg_tag
        """point-group tag."""

        so, so_n, ct_dict = _data_character[str(pg_tag)]
        s, a = _data_product_decomp[str(pg_tag)]

        # set irrep.
        self.__irrep = TagList.from_str(TagIrrep, ct_dict.keys())

        # symmetry operation
        self.__symmetry_operation = TagList.from_str(TagSymmetryOperation, so)
        self.__so_number = so_n

        # set character table.
        w = sp.symbols(r"\omega \omega^*")
        wpv = text_to_sympy("-1/2+sqrt(3)*I/2")
        wmv = text_to_sympy("-1/2-sqrt(3)*I/2")
        self.__character_symbol = {}
        for irrep, ct in ct_dict.items():
            self[TagIrrep(irrep)] = text_to_sympy(ct, local={"wp": wpv, "wm": wmv})
            self.__character_symbol[TagIrrep(irrep)] = text_to_sympy(ct, local={"wp": w[0], "wm": w[1]})

        # set parity conversion.
        cr = TagList.from_str(TagIrrep, _data_compatibility_relation[str(pg_tag)])
        n = len(cr) // 2
        cr_conv = cr[n:] + cr[:n]
        self.__parity_conv = {i: j for i, j in zip(cr, cr_conv)}

        # set product decomposition.
        self.__product_decomp_s = {}
        for ir1, p in zip(self.__irrep, s):
            for ir2, v in zip(self.__irrep, p):
                v = [(n, TagIrrep(ir)) for n, ir in v]
                self.__product_decomp_s[(ir1, ir2)] = v
        self.__product_decomp_a = {}
        for ir, v in zip(self.__irrep, a):
            v = [(n, TagIrrep(ir)) for n, ir in v]
            self.__product_decomp_a[ir] = v

    # ==================================================
    def __str__(self):
        return str(self.tag)

    # ==================================================
    def __repr__(self):
        return repr(self.tag)

    # ==================================================
    def latex(self):
        return self.tag.latex()

    # ==================================================
    def __getitem__(self, tag):
        if type(tag) == str:
            return self.get(TagIrrep(tag))
        else:
            return self.get(tag)

    # ==================================================
    def key_list(self):
        return TagList(self.keys())

    # ==================================================
    @property
    def irrep_list(self):
        """
        list of irreps.

        Returns:
            [TagIrrep]: list of irreps.
        Notes:
            - first element is identity irrep.
        """
        return self.__irrep

    # ==================================================
    def character(self, irrep=None, all_so=False, explicit=False):
        """
        character table.

        Args:
            irrep (str or TagIrrep, optional): irrep. tag ("" = identity irrep.).
            all_so (bool, optional): return characters for all symmetry operations ?
            explicit (bool, optional): use explicit expression of character ?

        Returns:
            - [sympy]: character table (symbol) for given irrep.
            or
            - {TagIrrep:[sympy]}: all irreps.
        """
        if irrep == "":
            irrep = self.irrep_list[0]
        ctbl = self if explicit else self.__character_symbol

        if irrep is None:
            if all_so:
                return {i: self._to_all(c) for i, c in ctbl.items()}
            else:
                return ctbl
        else:
            if type(irrep) == str:
                lst = list(map(str, self.irrep_list))
                assert irrep in lst, f"{irrep} is not find in {lst}"
                irrep = TagIrrep(irrep)
            if all_so:
                return self._to_all(ctbl[irrep])
            else:
                return ctbl[irrep]

    # ==================================================
    def _to_all(self, ch_cc):
        """
        convert character table for all symmetry operations.

        Args:
            ch_cc ([sympy]): characters for cc operations.

        Returns:
            [sympy]: characters for all symmetry operations.
        """
        ch = []
        for chi, n in zip(ch_cc, self.__so_number):
            ch.extend([chi] * n)

        return ch

    # ==================================================
    def symmetry_operation(self, ret_num=False):
        """
        representative symmetry operations in conjugacy class.

        Args:
            ret_num (bool, optional): return with the number of cc operations ?

        Returns:
            - [TagSymmetryOperation]: symmetry operation.
            or
            - [TagSymmetryOperation]: symmetry operations.
            - [int]: the number of symmetry operations in conjugacy class.
        """
        if ret_num:
            return self.__symmetry_operation, self.__so_number
        else:
            return self.__symmetry_operation

    # ==================================================
    def compatibility_relation(self, pg_tag):
        """
        compatibility_relation from the present point group to a given one.

        Args:
            pg_tag (TagGroup or str): point-group tag to obtain compatibility relation.

        Returns:
            {TagIrrep: [TagIrrep]}: compatibility relation from self to other.
        """
        if type(pg_tag) == str:
            other = TagGroup(pg_tag)
        else:
            other = pg_tag
        assert self.tag.subgroup == other.subgroup, "diffrent subgroup is given."
        p = self.irrep_list
        c = TagList.from_str(TagIrrep, _data_compatibility_relation[str(other)])
        return {i: j for i, j in zip(p, c)}

    # ==================================================
    def parity_conversion(self):
        """
        parity conversion.

        Returns:
            {TagIrrep: TagIrrep: correspondence from self to converted one.
        """
        return self.__parity_conv

    # ==================================================
    def symmetric_product_decomposition(self, irrep_pair, ret_ex=False):
        """
        symmetric product irreducible decomposition.

        Args:
            irrep_pair ((TagIrrep or str, TagIrrep or str)): a pair of irreps.
            ret_ex (bool, optional): return as an expression ?

        Returns:
            - [(int,TagIrrep)]: (the number of appearances, irrep.).
            or
            - sympy: decomposed expression.
        """
        ir1, ir2 = irrep_pair
        if type(ir1) == str:
            ir1 = TagIrrep(ir1)
        if type(ir2) == str:
            ir2 = TagIrrep(ir2)
        val = self.__product_decomp_s[(ir1, ir2)]

        if ret_ex:
            ex = 0
            for n, v in val:
                ex += n * v.symbol()
            val = ex

        return val

    # ==================================================
    def anti_symmetric_product_decomposition(self, irrep, ret_ex=False):
        """
        anti-symmetric product irreducible decomposition.

        Args:
            irrep (TagIrrep or str): irrep. tag.
            ret_ex (bool, optional): return as an expression ?

        Returns:
            - [(int,TagIrrep)]: (the number of appearances, irrep.).
            or
            - sympy: decomposed expression.
        """
        if type(irrep) == str:
            irrep = TagIrrep(irrep)
        val = self.__product_decomp_a[irrep]

        if ret_ex:
            ex = 0
            for n, v in val:
                ex += n * v.symbol()
            val = ex

        return val

    # ==================================================
    def irrep_decomposition(self, char):
        """
        irreducible decomposition.

        Args:
            char ([sympy]): characters to be decomposed for symmetry operations in conjugacy class, .

        Returns:
            [(int,TagIrrep)]: (the number of appearances, irrep.).
        """
        decomp = []
        g = sum(self.__so_number)
        for irrep, ir_char in self.items():
            s = 0
            for c, ic, d in zip(char, ir_char, self.__so_number):
                s += d * sp.conjugate(c) * ic
            n = sp.simplify(s) / g
            if n != 0:
                decomp.append((n, irrep))
        return decomp



"""
This file provides utility functions for multipole basis set.
"""
import sympy as sp
from gcoreutils.nsarray import NSArray


# ==================================================
def matrix_sum(ml):
    """
    sum of matrices.

    Args:
        ml (list): matrix [Matrix].

    Returns:
        Matrix: sum of matrices
    """
    if len(ml) == 0:
        return 0
    else:
        return sum(ml, NSArray.zeros(shape=ml[0].shape, style="matrix", fmt="sympy"))


# ==================================================
def matrix_to_dict(m):
    """
    convert matrix to list of matrix elements.

    Args:
        m (Matrix): matrix.

    Returns:
        tuple: tuple of matrix shape and elements, (shape, [(i,j,v)]).
    """
    m = NSArray(m, style="matrix", fmt="sympy")
    dim_r, dim_c = m.shape
    mat = []
    for i in range(dim_r):
        for j in range(dim_c):
            v = m[i, j]
            if v != 0:
                mat.append((i, j, str(v)))

    shape = (dim_r, dim_c)
    return shape, mat


# ==================================================
def dict_to_matrix(shape, elements):
    """
    convert dict to full matrix.

    Args:
        shape (tuple): matrix shape.
        elements (list): matrix elements, [(i,j,v)].

    Returns:
        NSArray: full matrix.
    """
    m = sp.zeros(shape[0], shape[1])
    for i, j, c in elements:
        m[i, j] = sp.sympify(c)

    fm = NSArray(str(m.tolist()), "matrix")

    return fm
